@@ -33,6 +33,8 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <dbus/dbus-glib.h>
+#include <dbus/dbus-glib-lowlevel.h>
+#include <dbus/dbus.h>
 
 #include "gupnp-network-manager.h"
 #include "gupnp-context.h"
@@ -75,9 +77,12 @@ typedef struct
 } NMDevice;
 
 struct _GUPnPNetworkManagerPrivate {
+        DBusConnection *dbus_connection;
         DBusGConnection *connection;
         DBusGProxy *manager_proxy;
 
+        GSource *idle_context_creation_src;
+
         GList *nm_devices;
 };
 
@@ -136,6 +141,8 @@ create_loopback_context (gpointer data)
         guint port;
         GError *error = NULL;
 
+        manager->priv->idle_context_creation_src = NULL;
+
         g_object_get (manager,
                       "main-context", &main_context,
                       "port", &port,
@@ -396,7 +403,6 @@ static void
 schedule_loopback_context_creation (GUPnPNetworkManager *manager)
 {
         GMainContext *main_context;
-        GSource *source;
 
         g_object_get (manager,
                       "main-context", &main_context,
@@ -405,30 +411,42 @@ schedule_loopback_context_creation (GUPnPNetworkManager *manager)
         /* Create contexts in mainloop so that is happens after user has hooked
          * to the "context-available" signal.
          */
-        source = g_idle_source_new ();
-        g_source_attach (source, main_context);
-        g_source_set_callback (source,
+        manager->priv->idle_context_creation_src = g_idle_source_new ();
+        g_source_attach (manager->priv->idle_context_creation_src,
+            main_context);
+        g_source_set_callback (manager->priv->idle_context_creation_src,
                                create_loopback_context,
                                manager,
                                NULL);
+        g_source_unref (manager->priv->idle_context_creation_src);
 }
 
 static void
 init_network_manager (GUPnPNetworkManager *manager)
 {
         GUPnPNetworkManagerPrivate *priv;
-        GError *error;
-        error = NULL;
+        DBusError derror;
+        GMainContext *main_context;
 
         priv = manager->priv;
 
-        priv->connection = dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);
-        if (priv->connection == NULL) {
-                g_warning ("Failed to connect to System Bus: %s",
-                           error->message);
+        g_object_get (manager, "main-context", &main_context, NULL);
+
+        /* Do fake open to initialize types */
+        dbus_g_connection_open ("", NULL);
+        dbus_error_init (&derror);
+        priv->dbus_connection = dbus_bus_get_private (DBUS_BUS_SYSTEM, &derror);
+        if (priv->dbus_connection == NULL) {
+                g_message ("Failed to connect to System Bus: %s",
+                           derror.message);
                 return;
         }
 
+        dbus_connection_setup_with_g_main (priv->dbus_connection, main_context);
+
+        priv->connection =
+            dbus_connection_get_g_connection (priv->dbus_connection);
+
         priv->manager_proxy = dbus_g_proxy_new_for_name (priv->connection,
                                                          DBUS_SERVICE_NM,
                                                          MANAGER_PATH,
@@ -497,10 +515,17 @@ gupnp_network_manager_constructed (GObject *object)
 static void
 gupnp_network_manager_dispose (GObject *object)
 {
+        GUPnPNetworkManager *manager;
         GUPnPNetworkManagerPrivate *priv;
         GObjectClass *object_class;
 
-        priv = GUPNP_NETWORK_MANAGER (object)->priv;
+        manager = GUPNP_NETWORK_MANAGER (object);
+        priv = manager->priv;
+
+        if (manager->priv->idle_context_creation_src) {
+                g_source_destroy (manager->priv->idle_context_creation_src);
+                manager->priv->idle_context_creation_src = NULL;
+        }
 
         if (priv->manager_proxy != NULL) {
                 g_object_unref (priv->manager_proxy);
@@ -513,6 +538,12 @@ gupnp_network_manager_dispose (GObject *object)
                 priv->nm_devices = NULL;
         }
 
+        if (priv->connection)  {
+          dbus_connection_close (priv->dbus_connection);
+          dbus_g_connection_unref (priv->connection);
+        }
+        priv->connection = NULL;
+
         /* Call super */
         object_class = G_OBJECT_CLASS (gupnp_network_manager_parent_class);
         object_class->dispose (object);
@@ -539,28 +570,33 @@ gupnp_network_manager_class_init (GUPnPNetworkManagerClass *klass)
 }
 
 gboolean
-gupnp_network_manager_is_available (void)
+gupnp_network_manager_is_available (GMainContext *main_context)
 {
-        DBusGConnection *connection;
+        DBusConnection *connection;
+        DBusError derror;
+        DBusGConnection *gconnection;
         DBusGProxy *dbus_proxy;
         GError *error = NULL;
         gboolean ret = FALSE;
 
-        connection = dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);
+        /* Do fake open to initialize types */
+        dbus_g_connection_open ("", NULL);
+        dbus_error_init (&derror);
+        connection = dbus_bus_get_private (DBUS_BUS_SYSTEM, &derror);
         if (connection == NULL) {
-                g_warning ("Failed to connect to System Bus: %s",
-                           error->message);
-                g_error_free (error);
-
-                return ret;
+                g_message ("Failed to connect to System Bus: %s",
+                           derror.message);
+                return FALSE;
         }
 
-        dbus_proxy = dbus_g_proxy_new_for_name (connection,
+        dbus_connection_setup_with_g_main (connection, main_context);
+        gconnection = dbus_connection_get_g_connection (connection);
+
+        dbus_proxy = dbus_g_proxy_new_for_name (gconnection,
                                                 DBUS_SERVICE_DBUS,
                                                 DBUS_PATH_DBUS,
                                                 DBUS_INTERFACE_DBUS);
 
-
         if (!dbus_g_proxy_call (dbus_proxy,
                                 "NameHasOwner",
                                 &error,
@@ -575,6 +611,8 @@ gupnp_network_manager_is_available (void)
         }
 
         g_object_unref (dbus_proxy);
+        dbus_connection_close (connection);
+        dbus_g_connection_unref (gconnection);
 
         return ret;
 }
