@@ -42,6 +42,8 @@
 #include "xml-util.h"
 #include "gvalue-util.h"
 
+#define SUBSCRIPTION_TIMEOUT 300 /* DLNA (7.2.22.1) enforced */
+
 G_DEFINE_TYPE (GUPnPService,
                gupnp_service,
                GUPNP_TYPE_SERVICE_INFO);
@@ -98,6 +100,9 @@ typedef struct {
                                            subscription */
 } SubscriptionData;
 
+static void
+send_initial_state (SubscriptionData *data);
+
 static SoupSession *
 gupnp_service_get_session (GUPnPService *service)
 {
@@ -179,26 +184,64 @@ notify_data_free (NotifyData *data)
 }
 
 struct _GUPnPServiceAction {
+        volatile gint ref_count;
+
         GUPnPContext *context;
 
         char         *name;
 
         SoupMessage  *msg;
 
-        xmlDoc       *doc;
+        GUPnPXMLDoc  *doc;
         xmlNode      *node;
 
         GString      *response_str;
 };
 
+GUPnPServiceAction *
+gupnp_service_action_ref (GUPnPServiceAction *action)
+{
+        g_return_val_if_fail (action, NULL);
+        g_return_val_if_fail (action->ref_count > 0, NULL);
+
+        g_atomic_int_inc (&action->ref_count);
+
+        return action;
+}
+
+void
+gupnp_service_action_unref (GUPnPServiceAction *action)
+{
+        g_return_if_fail (action);
+        g_return_if_fail (action->ref_count > 0);
+
+        if (g_atomic_int_dec_and_test (&action->ref_count)) {
+                g_free (action->name);
+                g_object_unref (action->msg);
+                g_object_unref (action->context);
+                g_object_unref (action->doc);
+
+                g_slice_free (GUPnPServiceAction, action);
+        }
+}
+
+/**
+ * gupnp_service_action_get_type:
+ *
+ * Get the gtype for GUPnPServiceActon
+ *
+ * Return value: The gtype of GUPnPServiceAction
+ **/
 GType
 gupnp_service_action_get_type (void)
 {
         static GType our_type = 0;
 
         if (our_type == 0)
-                our_type = g_pointer_type_register_static
-                                        ("GUPnPServiceAction");
+                our_type = g_boxed_type_register_static
+                        ("GUPnPServiceAction",
+                         (GBoxedCopyFunc) gupnp_service_action_ref,
+                         (GBoxedFreeFunc) gupnp_service_action_unref);
 
         return our_type;
 }
@@ -231,7 +274,7 @@ finalize_action (GUPnPServiceAction *action)
         response_body = g_string_free (action->response_str, FALSE);
 
         soup_message_set_response (action->msg,
-                                   "text/xml",
+                                   "text/xml; charset=\"utf-8\"",
                                    SOUP_MEMORY_TAKE,
                                    response_body,
                                    strlen (response_body));
@@ -248,16 +291,11 @@ finalize_action (GUPnPServiceAction *action)
         soup_server_unpause_message (server, action->msg);
 
         /* Cleanup */
-        g_free (action->name);
-        g_object_unref (action->msg);
-        g_object_unref (action->context);
-        xmlFreeDoc (action->doc);
-
-        g_slice_free (GUPnPServiceAction, action);
+        gupnp_service_action_unref (action);
 }
 
 /**
- * gupnp_service_action_get_name
+ * gupnp_service_action_get_name:
  * @action: A #GUPnPServiceAction
  *
  * Get the name of @action.
@@ -273,13 +311,14 @@ gupnp_service_action_get_name (GUPnPServiceAction *action)
 }
 
 /**
- * gupnp_service_action_get_locales
+ * gupnp_service_action_get_locales:
  * @action: A #GUPnPServiceAction
  *
  * Get an ordered (preferred first) #GList of locales preferred by
  * the client. Free list and elements after use.
  *
- * Return value: A #GList of #char* locale names.
+ * Return value: (element-type utf8) (transfer full): A #GList of #char*
+ * locale names.
  **/
 GList *
 gupnp_service_action_get_locales (GUPnPServiceAction *action)
@@ -290,7 +329,7 @@ gupnp_service_action_get_locales (GUPnPServiceAction *action)
 }
 
 /**
- * gupnp_service_action_get
+ * gupnp_service_action_get:
  * @action: A #GUPnPServiceAction
  * @Varargs: tuples of argument name, argument type, and argument value
  * location, terminated with %NULL.
@@ -311,7 +350,7 @@ gupnp_service_action_get (GUPnPServiceAction *action,
 }
 
 /**
- * gupnp_service_action_get_valist
+ * gupnp_service_action_get_valist:
  * @action: A #GUPnPServiceAction
  * @var_args: va_list of tuples of argument name, argument type, and argument
  * value location.
@@ -354,7 +393,50 @@ gupnp_service_action_get_valist (GUPnPServiceAction *action,
 }
 
 /**
- * gupnp_service_action_get_value
+ * gupnp_service_action_get_values:
+ * @action: A #GUPnPServiceAction
+ * @arg_names: (element-type utf8) : A #GList of argument names as string
+ * @arg_types: (element-type GType): A #GList of argument types as #GType
+ *
+ * A variant of #gupnp_service_action_get that uses #GList instead of varargs.
+ *
+ * Return value: (element-type GValue) (transfer full): The values as #GList of
+ * #GValue. #g_list_free the returned list and #g_value_unset and #g_slice_free
+ * each element.
+ **/
+GList *
+gupnp_service_action_get_values (GUPnPServiceAction *action,
+                                 GList              *arg_names,
+                                 GList              *arg_types)
+{
+        GList *arg_values;
+        int i;
+
+        g_return_val_if_fail (action != NULL, NULL);
+
+        arg_values = NULL;
+
+        for (i = 0; i < g_list_length (arg_names); i++) {
+                const char *arg_name;
+                GType arg_type;
+                GValue *arg_value;
+
+                arg_name = (const char *) g_list_nth_data (arg_names, i);
+                arg_type = (GType) g_list_nth_data (arg_types, i);
+
+                arg_value = g_slice_new0 (GValue);
+                g_value_init (arg_value, arg_type);
+
+                gupnp_service_action_get_value (action, arg_name, arg_value);
+
+                arg_values = g_list_append (arg_values, arg_value);
+        }
+
+        return arg_values;
+}
+
+/**
+ * gupnp_service_action_get_value: (skip)
  * @action: A #GUPnPServiceAction
  * @argument: The name of the argument to retrieve
  * @value: The #GValue to store the value of the argument, initialized
@@ -391,7 +473,36 @@ gupnp_service_action_get_value (GUPnPServiceAction *action,
 }
 
 /**
- * gupnp_service_action_set
+ * gupnp_service_action_get_gvalue:
+ * @action: A #GUPnPServiceAction
+ * @argument: The name of the argument to retrieve
+ * @type: The type of argument to retrieve
+ *
+ * Rename To: gupnp_service_action_get_value
+ * Retrieves the value of @argument into a GValue of type @type and returns it.
+ * The method exists only and only to satify PyGI, please use
+ * #gupnp_service_action_get_value and ignore this if possible.
+ *
+ * Return value: (transfer full): Value as #GValue associated with @action.
+ * #g_value_unset and #g_slice_free it after usage.
+ **/
+GValue *
+gupnp_service_action_get_gvalue (GUPnPServiceAction *action,
+                                       const char         *argument,
+                                       GType               type)
+{
+        GValue *val;
+
+        val = g_slice_new0 (GValue);
+        g_value_init (val, type);
+
+        gupnp_service_action_get_value (action, argument, val);
+
+        return val;
+}
+
+/**
+ * gupnp_service_action_set:
  * @action: A #GUPnPServiceAction
  * @Varargs: tuples of return value name, return value type, and
  * actual return value, terminated with %NULL.
@@ -412,7 +523,7 @@ gupnp_service_action_set (GUPnPServiceAction *action,
 }
 
 /**
- * gupnp_service_action_set_valist
+ * gupnp_service_action_set_valist:
  * @action: A #GUPnPServiceAction
  * @var_args: va_list of tuples of return value name, return value type, and
  * actual return value.
@@ -438,7 +549,8 @@ gupnp_service_action_set_valist (GUPnPServiceAction *action,
                 arg_type = va_arg (var_args, GType);
                 g_value_init (&value, arg_type);
 
-                G_VALUE_COLLECT (&value, var_args, 0, &collect_error);
+                G_VALUE_COLLECT (&value, var_args, G_VALUE_NOCOPY_CONTENTS,
+                                 &collect_error);
                 if (!collect_error) {
                         gupnp_service_action_set_value (action,
                                                         arg_name, &value);
@@ -457,7 +569,52 @@ gupnp_service_action_set_valist (GUPnPServiceAction *action,
 }
 
 /**
- * gupnp_service_action_set_value
+ * gupnp_service_action_set_values:
+ * @action: A #GUPnPServiceAction
+ * @arg_names: (element-type utf8) (transfer-none): A #GList of argument names
+ * @arg_values: (element-type GValue) (transfer-none): The #GList of values (as
+ * #GValues) that line up with @arg_names.
+ *
+ * Sets the specified action return values.
+ **/
+void
+gupnp_service_action_set_values (GUPnPServiceAction *action,
+                                 GList              *arg_names,
+                                 GList              *arg_values)
+{
+        g_return_if_fail (action != NULL);
+        g_return_if_fail (arg_names != NULL);
+        g_return_if_fail (arg_values != NULL);
+        g_return_if_fail (g_list_length (arg_names) ==
+                          g_list_length (arg_values));
+
+        if (action->msg->status_code == SOUP_STATUS_INTERNAL_SERVER_ERROR) {
+                g_warning ("Calling gupnp_service_action_set_value() after "
+                           "having called gupnp_service_action_return_error() "
+                           "is not allowed.");
+
+                return;
+        }
+
+        /* Append to response */
+        for (; arg_names; arg_names = arg_names->next) {
+                const char *arg_name;
+                GValue *value;
+
+                arg_name = arg_names->data;
+                value = arg_values->data;
+
+                xml_util_start_element (action->response_str, arg_name);
+                gvalue_util_value_append_to_xml_string (value,
+                                                        action->response_str);
+                xml_util_end_element (action->response_str, arg_name);
+
+                arg_values = arg_values->next;
+        }
+}
+
+/**
+ * gupnp_service_action_set_value:
  * @action: A #GUPnPServiceAction
  * @argument: The name of the return value to retrieve
  * @value: The #GValue to store the return value
@@ -488,7 +645,7 @@ gupnp_service_action_set_value (GUPnPServiceAction *action,
 }
 
 /**
- * gupnp_service_action_return
+ * gupnp_service_action_return:
  * @action: A #GUPnPServiceAction
  *
  * Return succesfully.
@@ -504,7 +661,7 @@ gupnp_service_action_return (GUPnPServiceAction *action)
 }
 
 /**
- * gupnp_service_action_return_error
+ * gupnp_service_action_return_error:
  * @action: A #GUPnPServiceAction
  * @error_code: The error code
  * @error_description: The error description, or %NULL if @error_code is
@@ -587,13 +744,14 @@ gupnp_service_action_return_error (GUPnPServiceAction *action,
 }
 
 /**
- * gupnp_service_action_get_message
+ * gupnp_service_action_get_message:
  * @action: A #GUPnPServiceAction
  *
  * Get the #SoupMessage associated with @action. Mainly intended for
  * applications to be able to read HTTP headers received from clients.
  *
- * Return value: #SoupMessage associated with @action. Unref after using it.
+ * Return value: (transfer full): #SoupMessage associated with @action. Unref
+ * after using it.
  **/
 SoupMessage *
 gupnp_service_action_get_message (GUPnPServiceAction *action)
@@ -782,9 +940,10 @@ control_server_handler (SoupServer        *server,
         /* Create action structure */
         action = g_slice_new (GUPnPServiceAction);
 
+        action->ref_count    = 1;
         action->name         = g_strdup (action_name);
         action->msg          = g_object_ref (msg);
-        action->doc          = doc;
+        action->doc          = gupnp_xml_doc_new(doc);
         action->node         = action_node;
         action->response_str = new_action_response_str (action_name,
                                                         soap_action);
@@ -884,39 +1043,58 @@ subscription_timeout (gpointer user_data)
         return FALSE;
 }
 
-/* Parse timeout header and return its value in seconds, or 0
- * if infinite */
-static int
-parse_and_limit_timeout (const char *timeout)
+static void
+send_initial_state (SubscriptionData *data)
 {
-        int timeout_seconds;
+        GQueue *queue;
+        char *mem;
+        GList *l;
 
-        timeout_seconds = 0;
+        /* Send initial event message */
+        queue = g_queue_new ();
 
-        if (strncmp (timeout, "Second-", strlen ("Second-")) == 0) {
-                /* We have a finite timeout */
-                timeout_seconds = CLAMP (atoi (timeout + strlen ("Second-")),
-                                         GENA_MIN_TIMEOUT,
-                                         GENA_MAX_TIMEOUT);
-        } else
-                timeout_seconds = GENA_MAX_TIMEOUT;
+        for (l = data->service->priv->state_variables; l; l = l->next) {
+                NotifyData *ndata;
 
-        return timeout_seconds;
+                ndata = g_slice_new0 (NotifyData);
+
+                g_signal_emit (data->service,
+                               signals[QUERY_VARIABLE],
+                               g_quark_from_string (l->data),
+                               l->data,
+                               &ndata->value);
+
+                if (!G_IS_VALUE (&ndata->value)) {
+                        g_slice_free (NotifyData, ndata);
+
+                        continue;
+                }
+
+                ndata->variable = g_strdup (l->data);
+
+                g_queue_push_tail (queue, ndata);
+        }
+
+        mem = create_property_set (queue);
+        notify_subscriber (data->sid, data, mem);
+
+        /* Cleanup */
+        g_queue_free (queue);
+
+        g_free (mem);
 }
 
+
 /* Subscription request */
 static void
 subscribe (GUPnPService *service,
            SoupMessage  *msg,
-           const char   *callback,
-           const char   *timeout)
+           const char   *callback)
 {
         SubscriptionData *data;
         char *start, *end, *uri;
-        int timeout_seconds;
-        GQueue *queue;
-        char *mem;
-        GList *l;
+        GUPnPContext *context;
+        GMainContext *main_context;
 
         data = g_slice_new0 (SubscriptionData);
 
@@ -952,26 +1130,17 @@ subscribe (GUPnPService *service,
         data->sid     = generate_sid ();
 
         /* Add timeout */
-        timeout_seconds = parse_and_limit_timeout (timeout);
-        if (timeout_seconds > 0) {
-                GUPnPContext *context;
-                GMainContext *main_context;
+	data->timeout_src = g_timeout_source_new_seconds (SUBSCRIPTION_TIMEOUT);
+	g_source_set_callback (data->timeout_src,
+                               subscription_timeout,
+                               data,
+                               NULL);
 
-	        data->timeout_src =
-                        g_timeout_source_new_seconds (timeout_seconds);
-	        g_source_set_callback (data->timeout_src,
-                                       subscription_timeout,
-                                       data,
-                                       NULL);
+        context = gupnp_service_info_get_context (GUPNP_SERVICE_INFO (service));
+        main_context = gssdp_client_get_main_context (GSSDP_CLIENT (context));
+        g_source_attach (data->timeout_src, main_context);
 
-                context = gupnp_service_info_get_context
-                        (GUPNP_SERVICE_INFO (service));
-                main_context = gssdp_client_get_main_context
-                        (GSSDP_CLIENT (context));
-	        g_source_attach (data->timeout_src, main_context);
-
-                g_source_unref (data->timeout_src);
-        }
+        g_source_unref (data->timeout_src);
 
         /* Add to hash */
         g_hash_table_insert (service->priv->subscriptions,
@@ -979,51 +1148,20 @@ subscribe (GUPnPService *service,
                              data);
 
         /* Respond */
-        subscription_response (service, msg, data->sid, timeout_seconds);
+        subscription_response (service, msg, data->sid, SUBSCRIPTION_TIMEOUT);
 
-        /* Send initial event message */
-        queue = g_queue_new ();
-
-        for (l = service->priv->state_variables; l; l = l->next) {
-                NotifyData *ndata;
-
-                ndata = g_slice_new0 (NotifyData);
-
-                g_signal_emit (service,
-                               signals[QUERY_VARIABLE],
-                               g_quark_from_string (l->data),
-                               l->data,
-                               &ndata->value);
-
-                if (!G_IS_VALUE (&ndata->value)) {
-                        g_slice_free (NotifyData, ndata);
-
-                        continue;
-                }
-
-                ndata->variable = g_strdup (l->data);
-
-                g_queue_push_tail (queue, ndata);
-        }
-
-        mem = create_property_set (queue);
-        notify_subscriber (data->sid, data, mem);
-
-        /* Cleanup */
-        g_queue_free (queue);
-
-        g_free (mem);
+        send_initial_state (data);
 }
 
 /* Resubscription request */
 static void
 resubscribe (GUPnPService *service,
              SoupMessage  *msg,
-             const char   *sid,
-             const char   *timeout)
+             const char   *sid)
 {
         SubscriptionData *data;
-        int timeout_seconds;
+        GUPnPContext *context;
+        GMainContext *main_context;
 
         data = g_hash_table_lookup (service->priv->subscriptions, sid);
         if (!data) {
@@ -1038,28 +1176,20 @@ resubscribe (GUPnPService *service,
                 data->timeout_src = NULL;
         }
 
-        timeout_seconds = parse_and_limit_timeout (timeout);
-        if (timeout_seconds > 0) {
-                GUPnPContext *context;
-                GMainContext *main_context;
+        data->timeout_src = g_timeout_source_new_seconds (SUBSCRIPTION_TIMEOUT);
+        g_source_set_callback (data->timeout_src,
+                               subscription_timeout,
+                               data,
+                               NULL);
 
-	        data->timeout_src = g_timeout_source_new_seconds
-                        (timeout_seconds);
-	        g_source_set_callback(data->timeout_src,
-				      subscription_timeout,
-				      data, NULL);
+        context = gupnp_service_info_get_context (GUPNP_SERVICE_INFO (service));
+        main_context = gssdp_client_get_main_context (GSSDP_CLIENT (context));
+        g_source_attach (data->timeout_src, main_context);
 
-                context = gupnp_service_info_get_context
-                        (GUPNP_SERVICE_INFO (service));
-                main_context = gssdp_client_get_main_context
-                        (GSSDP_CLIENT (context));
-	        g_source_attach (data->timeout_src, main_context);
-
-                g_source_unref (data->timeout_src);
-	}
+        g_source_unref (data->timeout_src);
 
         /* Respond */
-        subscription_response (service, msg, sid, timeout_seconds);
+        subscription_response (service, msg, sid, SUBSCRIPTION_TIMEOUT);
 }
 
 /* Unsubscription request */
@@ -1084,13 +1214,12 @@ subscription_server_handler (SoupServer        *server,
                              gpointer           user_data)
 {
         GUPnPService *service;
-        const char *callback, *nt, *timeout, *sid;
+        const char *callback, *nt, *sid;
 
         service = GUPNP_SERVICE (user_data);
 
         callback = soup_message_headers_get (msg->request_headers, "Callback");
         nt       = soup_message_headers_get (msg->request_headers, "NT");
-        timeout  = soup_message_headers_get (msg->request_headers, "Timeout");
         sid      = soup_message_headers_get (msg->request_headers, "SID");
 
         /* Choose appropriate handler */
@@ -1105,7 +1234,7 @@ subscription_server_handler (SoupServer        *server,
                                         (msg, SOUP_STATUS_PRECONDITION_FAILED);
 
                         } else {
-                                subscribe (service, msg, callback, timeout);
+                                subscribe (service, msg, callback);
 
                         }
 
@@ -1115,7 +1244,7 @@ subscription_server_handler (SoupServer        *server,
                                         (msg, SOUP_STATUS_BAD_REQUEST);
 
                         } else {
-                                resubscribe (service, msg, sid, timeout);
+                                resubscribe (service, msg, sid);
 
                         }
 
@@ -1148,41 +1277,23 @@ subscription_server_handler (SoupServer        *server,
         }
 }
 
-static GObject *
-gupnp_service_constructor (GType                  type,
-                           guint                  n_construct_params,
-                           GObjectConstructParam *construct_params)
+static void
+got_introspection (GUPnPServiceInfo          *info,
+                   GUPnPServiceIntrospection *introspection,
+                   const GError              *error,
+                   gpointer                   user_data)
 {
-        GObjectClass *object_class;
-        GObject *object;
-        GUPnPService *service;
-        GUPnPServiceInfo *info;
-        GError *error;
-        GUPnPServiceIntrospection *introspection;
+        GUPnPService *service = user_data;
         const GList *state_variables, *l;
-        GUPnPContext *context;
-        SoupServer *server;
-        SoupURI *uri;
-        char *url;
+        GHashTableIter iter;
+        gpointer data;
 
-        object_class = G_OBJECT_CLASS (gupnp_service_parent_class);
+        service = GUPNP_SERVICE (user_data);
 
-        /* Construct */
-        object = object_class->constructor (type,
-                                            n_construct_params,
-                                            construct_params);
-
-        service = GUPNP_SERVICE (object);
-        info    = GUPNP_SERVICE_INFO (object);
-
-        /* Get introspection and save state variable names */
-        error = NULL;
-
-        introspection = gupnp_service_info_get_introspection (info, &error);
         if (introspection) {
                 state_variables =
                         gupnp_service_introspection_list_state_variables
-                                                                (introspection);
+                                (introspection);
 
                 for (l = state_variables; l; l = l->next) {
                         GUPnPServiceStateVariableInfo *variable;
@@ -1198,13 +1309,61 @@ gupnp_service_constructor (GType                  type,
                 }
 
                 g_object_unref (introspection);
-        } else {
+        } else
                 g_warning ("Failed to get SCPD: %s\n"
                            "The initial event message will not be sent.",
-                           error->message);
+                           error ? error->message : "No error");
 
-                g_error_free (error);
-        }
+        g_hash_table_iter_init (&iter, service->priv->subscriptions);
+
+        while (g_hash_table_iter_next (&iter, NULL, &data))
+                send_initial_state ((SubscriptionData *) data);
+
+        g_object_unref (service);
+}
+
+static char *
+path_from_url (const char *url)
+{
+        SoupURI *uri;
+        gchar   *path;
+
+        uri = soup_uri_new (url);
+        path = soup_uri_to_string (uri, TRUE);
+        soup_uri_free (uri);
+
+        return path;
+}
+
+static GObject *
+gupnp_service_constructor (GType                  type,
+                           guint                  n_construct_params,
+                           GObjectConstructParam *construct_params)
+{
+        GObjectClass *object_class;
+        GObject *object;
+        GUPnPService *service;
+        GUPnPServiceInfo *info;
+        GUPnPContext *context;
+        SoupServer *server;
+        char *url;
+        char *path;
+
+        object_class = G_OBJECT_CLASS (gupnp_service_parent_class);
+
+        /* Construct */
+        object = object_class->constructor (type,
+                                            n_construct_params,
+                                            construct_params);
+
+        service = GUPNP_SERVICE (object);
+        info    = GUPNP_SERVICE_INFO (object);
+
+        /* Get introspection and save state variable names */
+        gupnp_service_info_get_introspection_async (info,
+                                                    got_introspection,
+                                                    object);
+        g_object_ref (object);
 
         /* Get server */
         context = gupnp_service_info_get_context (info);
@@ -1212,26 +1371,18 @@ gupnp_service_constructor (GType                  type,
 
         /* Run listener on controlURL */
         url = gupnp_service_info_get_control_url (info);
-        uri = soup_uri_new (url);
-        g_free (url);
-
-        url = soup_uri_to_string (uri, TRUE);
-        soup_uri_free (uri);
-
-        soup_server_add_handler (server, url,
+        path = path_from_url (url);
+        soup_server_add_handler (server, path,
                                  control_server_handler, object, NULL);
+        g_free (path);
         g_free (url);
 
         /* Run listener on eventSubscriptionURL */
         url = gupnp_service_info_get_event_subscription_url (info);
-        uri = soup_uri_new (url);
-        g_free (url);
-
-        url = soup_uri_to_string (uri, TRUE);
-        soup_uri_free (uri);
-
-        soup_server_add_handler (server, url,
+        path = path_from_url (url);
+        soup_server_add_handler (server, path,
                                  subscription_server_handler, object, NULL);
+        g_free (path);
         g_free (url);
 
         return object;
@@ -1274,11 +1425,15 @@ gupnp_service_set_property (GObject      *object,
         service = GUPNP_SERVICE (object);
 
         switch (property_id) {
-        case PROP_ROOT_DEVICE:
+        case PROP_ROOT_DEVICE: {
+                GUPnPRootDevice **dev;
+
                 service->priv->root_device = g_value_get_object (value);
+                dev = &(service->priv->root_device);
+
                 g_object_add_weak_pointer
                         (G_OBJECT (service->priv->root_device),
-                         (gpointer *) &service->priv->root_device);
+                         (gpointer *) dev);
 
                 service->priv->notify_available_id =
                         g_signal_connect_object (service->priv->root_device,
@@ -1289,6 +1444,7 @@ gupnp_service_set_property (GObject      *object,
                                                  0);
 
                 break;
+        }
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
                 break;
@@ -1320,10 +1476,36 @@ gupnp_service_dispose (GObject *object)
 {
         GUPnPService *service;
         GObjectClass *object_class;
+        GUPnPServiceInfo *info;
+        GUPnPContext *context;
+        SoupServer *server;
+        char *url;
+        char *path;
 
         service = GUPNP_SERVICE (object);
 
+        /* Get server */
+        info = GUPNP_SERVICE_INFO (service);
+        context = gupnp_service_info_get_context (info);
+        server = gupnp_context_get_server (context);
+
+        /* Remove listener on controlURL */
+        url = gupnp_service_info_get_control_url (info);
+        path = path_from_url (url);
+        soup_server_remove_handler (server, path);
+        g_free (path);
+        g_free (url);
+
+        /* Remove listener on eventSubscriptionURL */
+        url = gupnp_service_info_get_event_subscription_url (info);
+        path = path_from_url (url);
+        soup_server_remove_handler (server, path);
+        g_free (path);
+        g_free (url);
+
         if (service->priv->root_device) {
+                GUPnPRootDevice **dev = &(service->priv->root_device);
+
                 if (g_signal_handler_is_connected
                         (service->priv->root_device,
                          service->priv->notify_available_id)) {
@@ -1334,7 +1516,7 @@ gupnp_service_dispose (GObject *object)
 
                 g_object_remove_weak_pointer
                         (G_OBJECT (service->priv->root_device),
-                         (gpointer *) &service->priv->root_device);
+                         (gpointer *) dev);
 
                 service->priv->root_device = NULL;
         }
@@ -1402,7 +1584,7 @@ gupnp_service_class_init (GUPnPServiceClass *klass)
         g_type_class_add_private (klass, sizeof (GUPnPServicePrivate));
 
         /**
-         * GUPnPService:root-device
+         * GUPnPService:root-device:
          *
          * The containing #GUPnPRootDevice.
          **/
@@ -1420,7 +1602,7 @@ gupnp_service_class_init (GUPnPServiceClass *klass)
                                       G_PARAM_STATIC_BLURB));
 
         /**
-         * GUPnPService::action-invoked
+         * GUPnPService::action-invoked:
          * @service: The #GUPnPService that received the signal
          * @action: The invoked #GUPnPAction
          *
@@ -1436,13 +1618,13 @@ gupnp_service_class_init (GUPnPServiceClass *klass)
                                                action_invoked),
                               NULL,
                               NULL,
-                              g_cclosure_marshal_VOID__POINTER,
+                              g_cclosure_marshal_VOID__BOXED,
                               G_TYPE_NONE,
                               1,
                               GUPNP_TYPE_SERVICE_ACTION);
 
         /**
-         * GUPnPService::query-variable
+         * GUPnPService::query-variable:
          * @service: The #GUPnPService that received the signal
          * @variable: The variable that is being queried
          * @value: The location of the #GValue of the variable
@@ -1466,7 +1648,7 @@ gupnp_service_class_init (GUPnPServiceClass *klass)
                                                 is an outward argument! */);
 
         /**
-         * GUPnPService::notify-failed
+         * GUPnPService::notify-failed:
          * @service: The #GUPnPService that received the signal
          * @callback_url: The callback URL
          * @reason: A pointer to a #GError describing why the notify failed
@@ -1489,7 +1671,7 @@ gupnp_service_class_init (GUPnPServiceClass *klass)
 }
 
 /**
- * gupnp_service_notify
+ * gupnp_service_notify:
  * @service: A #GUPnPService
  * @Varargs: Tuples of variable name, variable type, and variable value,
  * terminated with %NULL.
@@ -1511,7 +1693,7 @@ gupnp_service_notify (GUPnPService *service,
 }
 
 /**
- * gupnp_service_notify_valist
+ * gupnp_service_notify_valist:
  * @service: A #GUPnPService
  * @var_args: A va_list of tuples of variable name, variable type, and variable
  * value, terminated with %NULL.
@@ -1537,7 +1719,8 @@ gupnp_service_notify_valist (GUPnPService *service,
                 var_type = va_arg (var_args, GType);
                 g_value_init (&value, var_type);
 
-                G_VALUE_COLLECT (&value, var_args, 0, &collect_error);
+                G_VALUE_COLLECT (&value, var_args, G_VALUE_NOCOPY_CONTENTS,
+                                 &collect_error);
                 if (!collect_error) {
                         gupnp_service_notify_value (service, var_name, &value);
 
@@ -1584,7 +1767,7 @@ notify_got_response (SoupSession *session,
                 /* Other failure: Try next callback or signal failure. */
                 if (data->callbacks->next) {
                         SoupURI *uri;
-                        SoupSession *session;
+                        SoupSession *service_session;
 
                         /* Call next callback */
                         data->callbacks = data->callbacks->next;
@@ -1597,9 +1780,9 @@ notify_got_response (SoupSession *session,
                         data->pending_messages = 
                                 g_list_prepend (data->pending_messages, msg);
 
-                        session = gupnp_service_get_session (data->service);
+                        service_session = gupnp_service_get_session (data->service);
 
-                        soup_session_requeue_message (session, msg);
+                        soup_session_requeue_message (service_session, msg);
                 } else {
                         /* Emit 'notify-failed' signal */
                         GError *error;
@@ -1702,22 +1885,21 @@ create_property_set (GQueue *queue)
         g_string_append (str,
                          "<?xml version=\"1.0\"?>"
                          "<e:propertyset xmlns:e="
-                                "\"urn:schemas-upnp-org:event-1-0\">"
-                         "<e:property>");
+                                "\"urn:schemas-upnp-org:event-1-0\">");
 
         /* Add variables */
         while ((data = g_queue_pop_head (queue))) {
+                xml_util_start_element (str, "e:property");
                 xml_util_start_element (str, data->variable);
                 gvalue_util_value_append_to_xml_string (&data->value, str);
                 xml_util_end_element (str, data->variable);
+                xml_util_end_element (str, "e:property");
 
                 /* Cleanup */
                 notify_data_free (data);
         }
 
-        g_string_append (str,
-                         "</e:property>"
-                         "</e:propertyset>");
+        g_string_append (str, "</e:propertyset>");
 
         /* Cleanup & return */
         return g_string_free (str, FALSE);
@@ -1742,7 +1924,7 @@ flush_notifications (GUPnPService *service)
 }
 
 /**
- * gupnp_service_notify_value
+ * gupnp_service_notify_value:
  * @service: A #GUPnPService
  * @variable: The name of the variable to notify
  * @value: The value of the variable
@@ -1754,13 +1936,13 @@ gupnp_service_notify_value (GUPnPService *service,
                             const char   *variable,
                             const GValue *value)
 {
+        NotifyData *data;
+
         g_return_if_fail (GUPNP_IS_SERVICE (service));
         g_return_if_fail (variable != NULL);
         g_return_if_fail (G_IS_VALUE (value));
 
         /* Queue */
-        NotifyData *data;
-
         data = g_slice_new0 (NotifyData);
 
         data->variable = g_strdup (variable);
@@ -1776,7 +1958,7 @@ gupnp_service_notify_value (GUPnPService *service,
 }
 
 /**
- * gupnp_service_freeze_notify
+ * gupnp_service_freeze_notify:
  * @service: A #GUPnPService
  *
  * Causes new notifications to be queued up until gupnp_service_thaw_notify()
@@ -1791,7 +1973,7 @@ gupnp_service_freeze_notify (GUPnPService *service)
 }
 
 /**
- * gupnp_service_thaw_notify
+ * gupnp_service_thaw_notify:
  * @service: A #GUPnPService
  *
  * Sends out any pending notifications, and stops queuing of new ones.
@@ -1814,7 +1996,7 @@ static char *
 strip_camel_case (char *camel_str)
 {
         char *stripped;
-        int i, j;
+        unsigned int i, j;
 
         /* Keep enough space for underscores */
         stripped = g_malloc (strlen (camel_str) * 2);
@@ -1928,7 +2110,7 @@ connect_names_to_signal_handlers (GUPnPService *service,
 }
 
 /**
- * gupnp_service_signals_autoconnect
+ * gupnp_service_signals_autoconnect:
  * @service: A #GUPnPService
  * @user_data: the data to pass to each of the callbacks
  * @error: return location for a #GError, or %NULL
